import os

def _norm(p: str) -> str:
    try:
        return os.path.abspath(os.path.realpath(p))
    except Exception:
        return os.path.abspath(p)

def find_idle_window_for_file(path: str, rows: list[dict]) -> dict | None:
    target = _norm(path)
    base   = os.path.basename(target)

    for w in rows:
        if not isinstance(w, dict):
            continue
        sig = w.get("program_signature") or {}
        if not isinstance(sig, dict):
            sig = {}
        args = sig.get("args") or []
        if not isinstance(args, (list, tuple)):
            args = []
        mod  = (sig.get("module") or "").lower()

        # Best case: IDLE via -m idlelib.idle with the exact file arg
        for a in args:
            try:
                if a.endswith(".py") and _norm(a) == target and mod == "idlelib.idle":
                    return w
            except Exception:
                pass

        # Fallback: window title heuristic
        title = (w.get("window_title") or "")
        if isinstance(title, str) and base.lower() in title.lower() and \
           os.path.dirname(target).lower() in title.lower():
            return w

    return None



from __future__ import annotations
from typing import List, Dict, Optional, Union, Callable, Any, Tuple
import os, re, time, subprocess
from .monitor_utils import *

# -------------------------
# /proc helpers (for signature)
# -------------------------
def _readlink_safe(path: str) -> Optional[str]:
    try:
        return os.readlink(path)
    except Exception:
        return None

def get_proc_exe(pid: Union[str, int]) -> Optional[str]:
    return _readlink_safe(f"/proc/{pid}/exe")

def get_proc_cwd(pid: Union[str, int]) -> Optional[str]:
    return _readlink_safe(f"/proc/{pid}/cwd")


def get_proc_cmdline(pid: Union[str, int]) -> List[str]:
    try:
        with open(f"/proc/{pid}/cmdline", "rb") as f:
            raw = f.read().split(b"\x00")
        return [x.decode("utf-8", "replace") for x in raw if x]
    except Exception:
        return []


def guess_python_entry_from_cmdline(args: List[str], cwd: Optional[str]) -> Dict[str, Optional[str]]:
    script_path = None
    module = None
    entry_kind = None

    if not args:
        return {"script_path": None, "module": None, "entry_kind": None, "args": []}

    i = 1
    while i < len(args) and args[i].startswith("-"):
        if args[i] == "-m" and i + 1 < len(args):
            module = args[i + 1]
            entry_kind = "module"
            break
        if args[i] == "-c":
            entry_kind = "inline"
            break
        i += 1

    # 🔍 NEW: scan the remainder of args for a .py file
    for a in args[i + 1:]:
        if a.endswith(".py"):
            script_path = os.path.normpath(os.path.join(cwd, a)) if cwd and not os.path.isabs(a) else a
            break

    if entry_kind is None and i < len(args):
        cand = args[i]
        cand_abs = os.path.normpath(os.path.join(cwd, cand)) if cwd and not os.path.isabs(cand) else cand
        if os.path.splitext(cand_abs)[1] in (".py", ".pyw", ""):
            script_path = cand_abs
            entry_kind = "script"

    return {
        "script_path": script_path,
        "module": module,
        "entry_kind": entry_kind,
        "args": args,
    }


def get_program_signature_for_pid(pid: Union[str, int]) -> Dict[str, Optional[str]]:
    exe = get_proc_exe(pid)
    cwd = get_proc_cwd(pid)
    argv = get_proc_cmdline(pid)
    py   = guess_python_entry_from_cmdline(argv, cwd) or {}
    return {
        'pid': str(pid),
        'exe': exe,
        'cwd': cwd,
        'argv': ' '.join(argv) if argv else None,
        'args': py.get('args') or [],          # <--- add this
        'script': py.get('script_path'),
        'module': py.get('module'),
        'kind': py.get('entry_kind'),
    }

# -------------------------
# wmctrl snapshot (fast)
# -------------------------
def _wmctrl_snapshot() -> List[Dict[str, str]]:
    """
    Single fast call: returns rows with id/desk/pid/host/class/title.
    We enrich lazily later.
    """
    out = subprocess.check_output(WMCTRL_LIST_CMD, text=True, errors="ignore")
    rows: List[Dict[str, str]] = []
    for line in out.splitlines():
        # wmctrl -lx -p columns: ID DESK PID HOST WM_CLASS TITLE
        parts = line.split(None, 5)
        if len(parts) < 6:
            continue
        wid, desk, pid, host, wmclass, title = parts
        rows.append({
            "window_id": wid,
            "desktop": desk,
            "pid": pid,
            "host": host,
            "wm_class": wmclass,     # e.g. "python3.Idle"
            "window_title": title,
        })
    return rows


# -------------------------
# enrich + parsing helpers
# -------------------------
def _enrich_row(row: Dict[str, str], *, with_signature: bool = False, with_geometry: bool = False) -> Dict[str, Any]:
    """Add optional fields lazily to a wmctrl row."""
    out: Dict[str, Any] = dict(row)
    if with_signature and row.get("pid"):
        out["program_signature"] = get_program_signature_for_pid(row["pid"])
    if with_geometry and row.get("window_id"):
        geom = get_window_geometry(row["window_id"])
        if geom:
            out["window_geometry"] = geom
            # map to a monitor if possible
            mon = get_monitor_for_xy(geom["x"], geom["y"])
            if mon:
                out.update(mon)
    return out

def get_all_parsed_windows(*, with_signature: bool = True, with_geometry: bool = False) -> List[Dict[str, Any]]:
    rows = _wmctrl_snapshot()
    return [_enrich_row(r, with_signature=with_signature, with_geometry=with_geometry) for r in rows]

# Backwards compatibility: your old functions
def get_windows_list() -> List[str]:
    """Kept for compatibility; prefer _wmctrl_snapshot() instead."""
    # Reconstruct a string like wmctrl -l -p (without class); only used if needed
    out = subprocess.check_output(["wmctrl", "-l", "-p"], text=True, errors="ignore")
    return out.splitlines()

def parse_window(window: str) -> Optional[Dict[str, Any]]:
    """Legacy parser for wmctrl -l -p lines; prefer get_all_parsed_windows()."""
    parts = window.split(None, 4)
    if len(parts) < 5:
        return None
    win_id, desktop, pid, host, title = parts
    row = {
        "window_id": win_id, "desktop": desktop, "pid": pid, "host": host,
        "wm_class": "", "window_title": title
    }
    # lazily enrich with signature; geometry only on demand
    return _enrich_row(row, with_signature=True, with_geometry=False)


# -------------------------
# actions
# -------------------------

def activate_window(window_id: str) -> bool:
    try:
        subprocess.run(['wmctrl', '-i', '-a', window_id], check=True, text=True)
        return True
    except subprocess.SubprocessError:
        return False

# -------------------------
# matchers (fast paths)
# -------------------------
def find_window_by_title_contains(substrings: List[str], rows: Optional[List[Dict[str, str]]] = None) -> Optional[Dict[str, Any]]:
    rows = rows or _wmctrl_snapshot()
    subs_l = [s.lower() for s in substrings]
    for r in rows:
        title = (r.get("window_title") or "").lower()
        if any(s in title for s in subs_l):
            return _enrich_row(r, with_signature=True, with_geometry=False)
    return None

def find_window_by_class(wm_class: str, rows: Optional[List[Dict[str, str]]] = None) -> Optional[Dict[str, Any]]:
    rows = rows or _wmctrl_snapshot()
    for r in rows:
        if r.get("wm_class") == wm_class:
            return _enrich_row(r, with_signature=True, with_geometry=False)
    return None

def find_window_for_script(script_path: str, rows: Optional[List[Dict[str, str]]] = None) -> Optional[Dict[str, Any]]:
    script_abs = os.path.abspath(script_path)
    
    rows = rows or _wmctrl_snapshot()
    for r in rows:
        pid = r.get("pid")
        if not pid:
            continue
        sig = get_program_signature_for_pid(pid)
        if sig.get("script") == script_abs:
            out = dict(r)
            out["program_signature"] = sig
            return out
    return None

def windows_matching_source(
    *,
    script_abs: Optional[str] = None,
    module: Optional[str] = None,
    exe_startswith: Optional[str] = None,
    cwd_abs: Optional[str] = None,
) -> List[Dict[str, Any]]:
    rows = _wmctrl_snapshot()
    matches: List[Dict[str, Any]] = []
    for r in rows:
        pid = r.get("pid")
        if not pid:
            continue
        sig = get_program_signature_for_pid(pid)
        ok = True
        if script_abs is not None and sig.get("script") != script_abs:
            ok = False
        if module is not None and sig.get("module") != module:
            ok = False
        if cwd_abs is not None and sig.get("cwd") != cwd_abs:
            ok = False
        if exe_startswith is not None:
            ex = sig.get("exe") or ""
            if not ex.startswith(exe_startswith):
                ok = False
        if ok:
            out = dict(r)
            out["program_signature"] = sig
            matches.append(out)
    return matches

# -------------------------
# convenience (what you used before)
# -------------------------
def get_parsed_windows() -> List[Dict[str, Any]]:
    """Compatibility shim for callers expecting a list of enriched windows."""
    return get_all_parsed_windows(with_signature=True, with_geometry=False)




from __future__ import annotations
from typing import List, Dict, Optional, Union, Callable
import threading,re,subprocess, time, shlex
import os

XRANDR_CMD          = ["xrandr", "--listmonitors"]  # faster to parse than full mode dump
WMCTRL_LIST_CMD     = ["wmctrl", "-lx", "-p"]       # includes class + pid
WMCTRL_MOVE_CMD     = ["wmctrl", "-i", "-r"]
XDOTOOL_GEOMETRY_CMD= ["xdotool", "getwindowgeometry"]

# --- regex ---
POSITION_REGEX = re.compile(r'Position:\s+(\d+),(\d+)')
GEOMETRY_REGEX = re.compile(r'Geometry:\s+(\d+)x(\d+)')

XRANDR_CMD     = ["xrandr"]  # or ["xrandr", "--listmonitors"] if you prefer
# Regex patterns

MONITOR_REGEX  = re.compile(
    r"""^
    (?P<name>\S+)\s+             # e.g. "DisplayPort-1-2"
    connected                     # literal word
    (?:\s+primary)?               # optional " primary"
    \s+
    (?P<width>\d+)x(?P<height>\d+)  # resolution, e.g. "2560x1440"
    \+(?P<x>\d+)\+(?P<y>\d+)        # offsets, e.g. "+0+0"
    """,
    re.VERBOSE
)
def get_monitors():
    monitors = []
    try:
        result = subprocess.run(
            XRANDR_CMD, capture_output=True, text=True, check=True
        )
        for line in result.stdout.splitlines():
            m = MONITOR_REGEX.search(line)
            if not m:
                continue
            gd = m.groupdict()
            monitors.append({
                "name":   gd["name"],
                "x":      int(gd["x"]),
                "y":      int(gd["y"]),
                "width":  int(gd["width"]),
                "height": int(gd["height"]),
            })
    except subprocess.SubprocessError as e:
        print(f"Error running xrandr: {e}")
    return monitors
def get_monitor_for_window(
    window_id: Optional[str] = None,
    x: Optional[int] = None,
    y: Optional[int] = None
) -> Dict[str, Union[str, int]]:
    """Determine which monitor a given window (or coordinate) is on."""

    if window_id and x is None and y is None:
        geom = get_window_geometry(window_id = window_id)
        if geom is None:
            return {}
        x, y = geom['x'], geom['y']

    if x is None or y is None:
        return {}
    for mon in get_monitors():
        if mon['x'] <= x < mon['x'] + mon['width'] and \
           mon['y'] <= y < mon['y'] + mon['height']:
            return {
                'monitor_name': mon['name'],
                'monitor_details': f"{mon['width']}x{mon['height']}+{mon['x']}+{mon['y']}",
                'win_x': x,
                'win_y': y
            }
    return {}
_MON_CACHE: Optional[List[Dict[str, int]]] = None
def get_mon_index(monitor_index: Union[int, str, None]) -> Optional[int]:
    if monitor_index is None:
        return None
    if isinstance(monitor_index, int):
        return monitor_index
    s = str(monitor_index)
    digits = ''.join(ch for ch in s if ch.isdigit())
    return int(digits) if digits else None

def get_monitors():
    monitors = []
    try:
        result = subprocess.run(
            XRANDR_CMD, capture_output=True, text=True, check=True
        )
        for line in result.stdout.splitlines():
            m = MONITOR_REGEX.search(line)
            if not m:
                continue
            gd = m.groupdict()
            monitors.append({
                "name":   gd["name"],
                "x":      int(gd["x"]),
                "y":      int(gd["y"]),
                "width":  int(gd["width"]),
                "height": int(gd["height"]),
            })
    except subprocess.SubprocessError as e:
        print(f"Error running xrandr: {e}")
    return monitors

def get_monitors_fast() -> List[Dict[str, int]]:
    """
    Cache of monitor geometries via `xrandr --listmonitors`.
    """
    global _MON_CACHE
    if _MON_CACHE is not None:
        return _MON_CACHE

    out = subprocess.run(
            XRANDR_CMD, capture_output=True, text=True, check=True
        )
    mons: List[Dict[str, int]] = []

    _MON_CACHE = get_monitors()
    return _MON_CACHE
def get_monitor_geom_by_index(idx: int) -> Optional[Dict[str, int]]:
    # Reuse your get_monitors()
    mons = get_monitors_fast()
    if idx < 0 or idx >= len(mons):
        return None
    m = mons[idx]
    return {"x": m["x"], "y": m["y"], "width": m["width"], "height": m["height"]}
def get_monitor_for_window(
    window_id: Optional[str] = None,
    x: Optional[int] = None,
    y: Optional[int] = None
) -> Dict[str, Union[str, int]]:
    """Determine which monitor a given window (or coordinate) is on."""

    if window_id and x is None and y is None:
        geom = get_window_geometry(window_id = window_id)
        if geom is None:
            return {}
        x, y = geom['x'], geom['y']

    if x is None or y is None:
        return {}
    for mon in get_monitors_fast():
        if mon['x'] <= x < mon['x'] + mon['width'] and \
           mon['y'] <= y < mon['y'] + mon['height']:
            return {
                'monitor_name': mon['name'],
                'monitor_details': f"{mon['width']}x{mon['height']}+{mon['x']}+{mon['y']}",
                'win_x': x,
                'win_y': y
            }
    return {}

def get_monitor_geom_by_index(idx: int) -> Optional[Dict[str, int]]:
    # Reuse your get_monitors()
    mons = get_monitors()
    if idx < 0 or idx >= len(mons):
        return None
    m = mons[idx]
    return {"x": m["x"], "y": m["y"], "width": m["width"], "height": m["height"]}
def move_window_to_monitor(window_id: str, monitor_index: int) -> bool:
    geom = get_monitor_geom_by_index(monitor_index)

    if not geom:
        print(f"[abstract_windows] monitor {monitor_index} not found.")
        return False
    x, y = geom["x"], geom["y"]
    try:
        subprocess.run(['wmctrl', '-i', '-r', window_id, '-e', f'0,{x},{y},-1,-1'],
                       check=True, text=True)
        return True
    except subprocess.SubprocessError as e:
        print(f"[abstract_windows] wmctrl move error: {e}")
        return False




from __future__ import annotations
from typing import List, Dict, Optional, Union
import subprocess, re, os

XRANDR_CMD = ["xrandr"]
WMCTRL_MOVE_CMD = ["wmctrl", "-i", "-r"]

# Monitor info cache
_MON_CACHE: Optional[List[Dict[str, int]]] = None

MONITOR_REGEX = re.compile(
    r"""^
    (?P<name>\S+)\s+connected(?:\s+primary)?\s+
    (?P<width>\d+)x(?P<height>\d+)\+(?P<x>\d+)\+(?P<y>\d+)
    """,
    re.VERBOSE
)

def _run_xrandr() -> str:
    try:
        return subprocess.check_output(XRANDR_CMD, text=True, errors="ignore")
    except Exception:
        return ""

def _parse_xrandr_output(out: str) -> List[Dict[str, int]]:
    mons = []
    for line in out.splitlines():
        m = MONITOR_REGEX.search(line)
        if not m:
            continue
        gd = m.groupdict()
        mons.append({k: int(v) if v.isdigit() else v for k, v in gd.items()})
    return mons

def get_monitors() -> List[Dict[str, int]]:
    """Fetch all connected monitors (no cache)."""
    return _parse_xrandr_output(_run_xrandr())

def get_monitors_fast() -> List[Dict[str, int]]:
    """Cached version of get_monitors()."""
    global _MON_CACHE
    if _MON_CACHE is None:
        _MON_CACHE = get_monitors()
    return _MON_CACHE

def get_monitor_geom_by_index(idx: int) -> Optional[Dict[str, int]]:
    mons = get_monitors_fast()
    if 0 <= idx < len(mons):
        return mons[idx]
    return None

def get_monitor_for_xy(x: int, y: int) -> Dict[str, Union[str, int]]:
    """Return which monitor contains coordinates (x, y)."""
    for mon in get_monitors_fast():
        if mon["x"] <= x < mon["x"] + mon["width"] and mon["y"] <= y < mon["y"] + mon["height"]:
            return {
                "monitor_name": mon["name"],
                "monitor_details": f'{mon["width"]}x{mon["height"]}+{mon["x"]}+{mon["y"]}',
                "win_x": x,
                "win_y": y,
            }
    return {}

def get_mon_index(monitor_index: Union[int, str, None]) -> Optional[int]:
    if monitor_index is None:
        return None
    if isinstance(monitor_index, int):
        return monitor_index
    digits = "".join(ch for ch in str(monitor_index) if ch.isdigit())
    return int(digits) if digits else None

def move_window_to_monitor(window_id: str, monitor_index: int) -> bool:
    geom = get_monitor_geom_by_index(monitor_index)
    if not geom:
        print(f"[abstract_windows] monitor {monitor_index} not found.")
        return False
    try:
        subprocess.run(
            WMCTRL_MOVE_CMD + [window_id, "-e", f"0,{geom['x']},{geom['y']},-1,-1"],
            check=True, text=True
        )
        return True
    except subprocess.SubprocessError as e:
        print(f"[abstract_windows] wmctrl move error: {e}")
        return False


# -------------------------
# enrich + parsing helpers
# -------------------------
def _enrich_row(row: Dict[str, str], *, with_signature: bool = False, with_geometry: bool = False) -> Dict[str, Any]:
    """Add optional fields lazily to a wmctrl row."""
    out: Dict[str, Any] = dict(row)
    if with_signature and row.get("pid"):
        out["program_signature"] = get_program_signature_for_pid(row["pid"])
    if with_geometry and row.get("window_id"):
        geom = get_window_geometry(row["window_id"])
        if geom:
            out["window_geometry"] = geom
            # map to a monitor if possible
            mon = get_monitor_for_xy(geom["x"], geom["y"])
            if mon:
                out.update(mon)
    return out

def get_all_parsed_windows(*, with_signature: bool = True, with_geometry: bool = False) -> List[Dict[str, Any]]:
    rows = _wmctrl_snapshot()
    return [_enrich_row(r, with_signature=with_signature, with_geometry=with_geometry) for r in rows]

# Backwards compatibility: your old functions
def get_windows_list() -> List[str]:
    """Kept for compatibility; prefer _wmctrl_snapshot() instead."""
    # Reconstruct a string like wmctrl -l -p (without class); only used if needed
    out = subprocess.check_output(["wmctrl", "-l", "-p"], text=True, errors="ignore")
    return out.splitlines()

def parse_window(window: str) -> Optional[Dict[str, Any]]:
    """Legacy parser for wmctrl -l -p lines; prefer get_all_parsed_windows()."""
    parts = window.split(None, 4)
    if len(parts) < 5:
        return None
    win_id, desktop, pid, host, title = parts
    row = {
        "window_id": win_id, "desktop": desktop, "pid": pid, "host": host,
        "wm_class": "", "window_title": title
    }
    # lazily enrich with signature; geometry only on demand
    return _enrich_row(row, with_signature=True, with_geometry=False)



from __future__ import annotations
import os, time, json, subprocess
from typing import List, Dict, Optional, Union, Any
from .window_utils import (  # your module
    get_windows_list, parse_window, move_window_to_monitor, activate_window,get_all_parsed_windows,
    find_window_for_script
)
from .monitor_utils import get_mon_index
# ----------------- small helpers -----------------

def get_window_ids(parsed_windows: Optional[List[Dict[str, Any]]] = None) -> set[str]:
    parsed_windows = parsed_windows or get_all_parsed_windows()
    return {w.get('window_id') for w in parsed_windows if w.get('window_id')}



def find_window_by_title_fragments(fragments: List[str]) -> Optional[Dict[str, Any]]:
    frags = [f.lower() for f in fragments]
    for w in get_all_parsed_windows():
        title = (w.get('window_title') or '').lower()
        if any(f in title for f in frags):
            return w
    return None

def _place_and_focus(w: Dict[str, Any], monitor_index: Optional[int]) -> Dict[str, Union[str, bool]]:
    wid = w.get('window_id')
    moved = True
    if wid and monitor_index is not None:
        moved = move_window_to_monitor(wid, monitor_index)
    activated = bool(wid) and activate_window(wid)
    return {"launched": False, "window_id": wid or "", "moved": moved, "activated": activated}

# ----------------- core: safe, single-instance -----------------

def get_new_window_info(
    launch_cmd: List[str],
    cwd: Optional[str],
    *,
    timeout: float = 10.0,
    poll_interval: float = 0.25,
    prefer_pid_match: bool = True,
    title_fallback: Optional[List[str]] = None,
) -> Optional[Dict[str, Any]]:
    """
    Launch once, then wait for a new window. Don't spin; obey timeout.
    Prefer matching window by spawned PID; optionally fall back to title.
    """
    before_ids = get_window_ids()
    proc = subprocess.Popen(launch_cmd, cwd=cwd)
    deadline = time.time() + timeout

    last_seen: Optional[Dict[str, Any]] = None
    while time.time() < deadline:
        time.sleep(poll_interval)
        parsed = get_all_parsed_windows()

        # 1) Prefer exact new ID (set difference)
        for w in parsed:
            wid = w.get('window_id')
            if wid and wid not in before_ids:
                last_seen = w
                # if WM gave PID, try PID equality
                if prefer_pid_match and w.get("pid") == str(proc.pid):
                    return w

        # 2) Exact PID match (even if ID reused)
        for w in parsed:
            if w.get("pid") == str(proc.pid):
                return w

        # 3) Title fallback (helps WMs that delay PID)
        if title_fallback:
            maybe = find_window_by_title_fragments(title_fallback)
            if maybe:
                last_seen = maybe

    return last_seen  # may be None

def ensure_single_instance_or_launch(
    *,
    path: str,
    match_strings: Optional[List[str]] = None,
    monitor_index: Union[int, str, None] = 1,
    launch_cmd: List[str],
    cwd: Optional[str] = None,
    appear_timeout_sec: float = 7.0,
    poll_interval_sec: float = 0.12,
    lock_path: Optional[str] = None,
    debounce_sec: float = 0.75,
) -> Dict[str, Union[str, bool]]:
    """
    Safe single-instance:
      1) If a window for this script is open -> place & focus.
      2) Else launch once; wait (with timeout) for its window; place & focus.

    Includes:
      - lock/debounce to avoid double-click storms
      - no infinite loops
    """
    # ---- debounce: stop click-storms from spawning multiple processes ----
    if lock_path is None:
        lock_path = os.path.expanduser("~/.cache/abstract-ide/launch.lock")
    os.makedirs(os.path.dirname(lock_path), exist_ok=True)

    now = time.time()
    try:
        if os.path.exists(lock_path):
            with open(lock_path, "r") as f:
                data = json.load(f)
            last = float(data.get("ts", 0))
            if now - last < debounce_sec:
                # Another launcher just ran; try to bring to front instead of relaunching.
                existing = find_window_for_script(path)
                if existing:
                    return _place_and_focus(existing, get_mon_index(monitor_index))
        # update stamp
        with open(lock_path, "w") as f:
            json.dump({"ts": now}, f)
    except Exception:
        pass  # don't fail launching because of lock i/o

    # ---- 1) check for existing window by exact source signature ----
    existing = find_window_for_script(path)
    if existing:
        return _place_and_focus(existing, get_mon_index(monitor_index))

    # ---- 2) launch once, then locate safely (PID / new ID / title fallback) ----
    win = get_new_window_info(
        launch_cmd, cwd,
        timeout=appear_timeout_sec,
        poll_interval=poll_interval_sec,
        prefer_pid_match=True,
        title_fallback=(match_strings or []),
    )
    if not win:
        return {"launched": True, "window_id": "", "moved": False, "activated": False}

    wid = win.get("window_id") or ""
    moved = True
    mon_idx = get_mon_index(monitor_index)
    if mon_idx is not None:
        moved = move_window_to_monitor(wid, mon_idx)
    activated = activate_window(wid)
    return {"launched": True, "window_id": wid, "moved": moved, "activated": activated}

# ----------------- convenience: launch python with conda -----------------

def launch_python_conda_script(
    path: str,
    *,
    env_name: str = "base",
    conda_exe: str = "/home/computron/miniconda/bin/conda",
    display: str = ":0",
    monitor_index: Union[int, str, None] = 1,
) -> Dict[str, Union[str, bool]]:
    """
    Launch a python script via `conda run` (no shell sourcing), single-instance & safe.
    """
    script_abs = os.path.abspath(path)
    workdir = os.path.dirname(script_abs)

    launch_cmd = [
        conda_exe, "run", "-n", env_name, "--no-capture-output",
        "env", f"DISPLAY={display}",
        "python", script_abs,
    ]

    match_titles = [os.path.basename(script_abs)]  # light fallback
    return ensure_single_instance_or_launch(
        path=script_abs,
        match_strings=match_titles,
        monitor_index=monitor_index,   # ← integer like 1 (NOT DISPLAY)
        launch_cmd=launch_cmd,
        cwd=workdir,
        appear_timeout_sec=8.0,
        poll_interval_sec=0.15,
    )
def edit_python_conda_script(
    path: str,
    *,
    env_name: str = "base",
    conda_exe: str = "/home/computron/miniconda/bin/conda",
    display: str = ":0",
    monitor_index: Union[int, str, None] = 1,
) -> Dict[str, Union[str, bool]]:
    """
    Open the given Python file in IDLE (edit mode) using the given Conda env,
    but if that file is already open in any IDLE editor window,
    just focus that window instead of reopening it.
    """

    from .idle_utils import find_idle_window_for_file  # ensures edit-window match
    script_abs = os.path.abspath(path)
    workdir = os.path.dirname(script_abs)

    # collect all currently open windows
    parsed = get_all_parsed_windows(with_signature=True)
    existing = find_idle_window_for_file(script_abs, parsed)

    if existing:
        # ✅ bring to front instead of reopening
        wid = existing.get("window_id")
        print(f"[abstract_windows] focusing existing IDLE window for {script_abs}")
        return _place_and_focus(existing, get_mon_index(monitor_index))

    # --- no existing IDLE window found; open it once ---
    match_titles = [os.path.basename(script_abs), "IDLE"]

    launch_cmd = [
        conda_exe, "run", "-n", env_name, "--no-capture-output",
        "env", f"DISPLAY={display}",
        "python", "-m", "idlelib.idle", "-e", script_abs,
    ]

    print(f"[abstract_windows] launching IDLE edit for {script_abs}")
    return ensure_single_instance_or_launch(
        path=script_abs,
        match_strings=match_titles,
        monitor_index=monitor_index,
        launch_cmd=launch_cmd,
        cwd=workdir,
        appear_timeout_sec=10.0,
        poll_interval_sec=0.25,
    )




from __future__ import annotations
from typing import List, Dict, Optional, Any
import subprocess, os, platform, time, re
from contextlib import suppress

WMCTRL_LIST_CMD = ["wmctrl", "-lx", "-p"]

# Cached wmctrl results (avoid repeated subprocess hits)
_WMCTRL_CACHE = {"t": 0.0, "rows": []}


# ----------------------------
# /proc helpers (Linux only)
# ----------------------------
def _readlink_safe(path: str) -> Optional[str]:
    with suppress(Exception):
        return os.readlink(path)
    return None

def get_proc_exe(pid: str | int) -> Optional[str]:
    return _readlink_safe(f"/proc/{pid}/exe")

def get_proc_cwd(pid: str | int) -> Optional[str]:
    return _readlink_safe(f"/proc/{pid}/cwd")

def get_proc_cmdline(pid: str | int) -> List[str]:
    try:
        with open(f"/proc/{pid}/cmdline", "rb") as f:
            raw = f.read().split(b"\x00")
        return [x.decode("utf-8", "replace") for x in raw if x]
    except Exception:
        return []


# ----------------------------
# Window Snapshot + Enrichment
# ----------------------------
def _wmctrl_snapshot() -> List[Dict[str, str]]:
    """Returns parsed wmctrl -lx -p output with caching (0.5s TTL)."""
    now = time.time()
    if now - _WMCTRL_CACHE["t"] < 0.5:
        return _WMCTRL_CACHE["rows"]

    try:
        out = subprocess.check_output(WMCTRL_LIST_CMD, text=True, errors="ignore")
    except subprocess.SubprocessError:
        return []

    rows = []
    for line in out.splitlines():
        parts = line.split(None, 5)
        if len(parts) == 6:
            wid, desk, pid, host, wmclass, title = parts
            rows.append({
                "window_id": wid,
                "desktop": desk,
                "pid": pid,
                "host": host,
                "wm_class": wmclass,
                "window_title": title,
            })
    _WMCTRL_CACHE.update({"t": now, "rows": rows})
    return rows


# ----------------------------
# Focusing Windows (cross-platform)
# ----------------------------
def _focus_linux(title_substring: str) -> bool:
    """Try to focus a window by title on Linux (supports wmctrl, xdotool, labwc)."""
    title = title_substring.lower()
    # wmctrl path
    try:
        windows = subprocess.check_output(["wmctrl", "-l"], text=True, errors="ignore")
        for line in windows.splitlines():
            if title in line.lower():
                win_id = line.split()[0]
                subprocess.run(["wmctrl", "-i", "-a", win_id])
                return True
    except Exception:
        pass
    # xdotool fallback
    with suppress(Exception):
        subprocess.run(["xdotool", "search", "--name", title_substring, "windowactivate"])
        return True
    # labwc fallback
    if os.environ.get("XDG_SESSION_DESKTOP", "").lower().startswith("labwc"):
        with suppress(Exception):
            subprocess.run(["labwc-client", "focus", title_substring])
            return True
    return False


def _focus_macos(app_or_title: str) -> bool:
    """Focus a macOS window by app name or title."""
    with suppress(Exception):
        subprocess.run(["osascript", "-e", f'tell application "{app_or_title}" to activate'])
        return True
    return False


def _focus_windows(title_substring: str) -> bool:
    """Focus a window by title using pygetwindow on Windows."""
    with suppress(Exception):
        import pygetwindow as gw
        wins = gw.getWindowsWithTitle(title_substring)
        if wins:
            win = wins[0]
            win.restore()
            win.activate()
            return True
    return False


def focus_window(title_substring: str) -> bool:
    """Cross-platform focus API."""
    system = platform.system()
    if system == "Linux":
        return _focus_linux(title_substring)
    elif system == "Darwin":
        return _focus_macos(title_substring)
    elif system == "Windows":
        return _focus_windows(title_substring)
    return False


# ----------------------------
# Activation (by window_id)
# ----------------------------
def activate_window(window_id: str) -> bool:
    """Activate a specific window by ID (Linux/X11 only)."""
    try:
        subprocess.run(["wmctrl", "-i", "-a", window_id], check=True, text=True)
        return True
    except subprocess.SubprocessError:
        return False

def guess_python_entry_from_cmdline(args: List[str], cwd: Optional[str]) -> Dict[str, Optional[str]]:
    script_path = None
    module = None
    entry_kind = None

    if not args:
        return {"script_path": None, "module": None, "entry_kind": None, "args": []}

    i = 1
    while i < len(args) and args[i].startswith("-"):
        if args[i] == "-m" and i + 1 < len(args):
            module = args[i + 1]
            entry_kind = "module"
            break
        if args[i] == "-c":
            entry_kind = "inline"
            break
        i += 1

    # 🔍 NEW: scan the remainder of args for a .py file
    for a in args[i + 1:]:
        if a.endswith(".py"):
            script_path = os.path.normpath(os.path.join(cwd, a)) if cwd and not os.path.isabs(a) else a
            break

    if entry_kind is None and i < len(args):
        cand = args[i]
        cand_abs = os.path.normpath(os.path.join(cwd, cand)) if cwd and not os.path.isabs(cand) else cand
        if os.path.splitext(cand_abs)[1] in (".py", ".pyw", ""):
            script_path = cand_abs
            entry_kind = "script"

    return {
        "script_path": script_path,
        "module": module,
        "entry_kind": entry_kind,
        "args": args,
    }


# ----------------------------
# Finders
# ----------------------------
def find_window_by_title_contains(
    substrings: List[str],
    rows: Optional[List[Dict[str, str]]] = None
) -> Optional[Dict[str, Any]]:
    rows = rows or _wmctrl_snapshot()
    subs_l = [s.lower() for s in substrings]
    for r in rows:
        title = (r.get("window_title") or "").lower()
        if any(s in title for s in subs_l):
            return r
    return None
def find_window_for_script(script_path: str, rows: Optional[List[Dict[str, str]]] = None) -> Optional[Dict[str, Any]]:
    script_abs = os.path.abspath(script_path)
    
    rows = rows or _wmctrl_snapshot()
    for r in rows:
        pid = r.get("pid")
        if not pid:
            continue
        sig = get_program_signature_for_pid(pid)
        if sig.get("script") == script_abs:
            out = dict(r)
            out["program_signature"] = sig
            return out
    return None
def get_program_signature_for_pid(pid: Union[str, int]) -> Dict[str, Optional[str]]:
    exe = get_proc_exe(pid)
    cwd = get_proc_cwd(pid)
    argv = get_proc_cmdline(pid)
    py   = guess_python_entry_from_cmdline(argv, cwd) or {}
    return {
        'pid': str(pid),
        'exe': exe,
        'cwd': cwd,
        'argv': ' '.join(argv) if argv else None,
        'args': py.get('args') or [],          # <--- add this
        'script': py.get('script_path'),
        'module': py.get('module'),
        'kind': py.get('entry_kind'),
    }

# -------------------------
# enrich + parsing helpers
# -------------------------
def _enrich_row(row: Dict[str, str], *, with_signature: bool = False, with_geometry: bool = False) -> Dict[str, Any]:
    """Add optional fields lazily to a wmctrl row."""
    out: Dict[str, Any] = dict(row)
    if with_signature and row.get("pid"):
        out["program_signature"] = get_program_signature_for_pid(row["pid"])
    if with_geometry and row.get("window_id"):
        geom = get_window_geometry(row["window_id"])
        if geom:
            out["window_geometry"] = geom
            # map to a monitor if possible
            mon = get_monitor_for_xy(geom["x"], geom["y"])
            if mon:
                out.update(mon)
    return out

def get_all_parsed_windows(*, with_signature: bool = True, with_geometry: bool = False) -> List[Dict[str, Any]]:
    rows = _wmctrl_snapshot()
    return [_enrich_row(r, with_signature=with_signature, with_geometry=with_geometry) for r in rows]

# Backwards compatibility: your old functions
def get_windows_list() -> List[str]:
    """Kept for compatibility; prefer _wmctrl_snapshot() instead."""
    # Reconstruct a string like wmctrl -l -p (without class); only used if needed
    out = subprocess.check_output(["wmctrl", "-l", "-p"], text=True, errors="ignore")
    return out.splitlines()

def parse_window(window: str) -> Optional[Dict[str, Any]]:
    """Legacy parser for wmctrl -l -p lines; prefer get_all_parsed_windows()."""
    parts = window.split(None, 4)
    if len(parts) < 5:
        return None
    win_id, desktop, pid, host, title = parts
    row = {
        "window_id": win_id, "desktop": desktop, "pid": pid, "host": host,
        "wm_class": "", "window_title": title
    }
    # lazily enrich with signature; geometry only on demand
    return _enrich_row(row, with_signature=True, with_geometry=False)



import subprocess
import time
from typing import List
from ..string_comp import *
from ..window_utils import *
# tweak if your browser uses Page_Down instead of Tab
NEXT_TAB_KEY = "ctrl+Tab"
# how many times to press before giving up
MAX_TABS = 50
# time for browser to update title
DELAY = 0.1
# Browser title signatures
BROWSER_PATTERNS = [
    'Mozilla', 'Firefox', 'Chrome', 'Chromium',
    'Microsoft Edge', 'Opera', 'Safari'
]
def is_browser(win):
    if get_strings_in_string(win.get('window_title'),BROWSER_PATTERNS):
        return True
    return False

def get_tab_titles(win):
    win_id = win["window_id"]
    tab_titles = cycle_tabs(win_id)
    win['tab_titles'] = tab_titles
    return win

def get_browsers(parsed_windows=None,active_windows=None):
    parsed_windows = parsed_windows or get_parsed_windows()
    browsers = [get_tab_titles(parsed_window) for parsed_window in parsed_windows if is_browser(parsed_window)]
    return browsers

def activate_window(win_id: str):
    subprocess.run(["xdotool", "windowactivate", "--sync", win_id], check=True)

def get_window_title(win_id: str) -> str:
    res = subprocess.run(
        ["xdotool", "getwindowname", win_id],
        capture_output=True, text=True, check=True
    )
    return res.stdout.strip()

def switch_tab(win_id: str):
    # sends the next‐tab key only to that window
    subprocess.run(
        ["xdotool", "key", "--window", win_id, NEXT_TAB_KEY],
        check=True
    )

def cycle_tabs(win_id: str,window_name=None) -> List[str]:
    """Returns the ordered list of tab-titles in the given window."""
    activate_window(win_id)
    first_title = get_window_title(win_id)
    titles = [first_title]

    for _ in range(MAX_TABS):
        switch_tab(win_id)
        time.sleep(DELAY)
        title = get_window_title(win_id)
        if window_name and title == window_name:
            return 
        if title == first_title:
            break
        titles.append(title)

    return titles



from __future__ import annotations
import os, time, json, subprocess
from typing import List, Dict, Optional, Union, Any
from .monitor_utils import move_window_to_monitor, get_mon_index
# ✅ fixed import block
from .window_core import *
def _place_and_focus(w: Dict[str, Any], monitor_index: Optional[int]) -> Dict[str, Union[str, bool]]:
    wid = w.get("window_id")
    moved = True
    if wid and monitor_index is not None:
        moved = move_window_to_monitor(wid, monitor_index)
    activated = bool(wid) and activate_window(wid)
    if not activated and w.get("window_title"):
        focus_window(w["window_title"])
    return {"launched": False, "window_id": wid or "", "moved": moved, "activated": activated}

# ----------------- small helpers -----------------

def get_window_ids(parsed_windows: Optional[List[Dict[str, Any]]] = None) -> set[str]:
    parsed_windows = parsed_windows or get_all_parsed_windows()
    return {w.get('window_id') for w in parsed_windows if w.get('window_id')}



def find_window_by_title_fragments(fragments: List[str]) -> Optional[Dict[str, Any]]:
    frags = [f.lower() for f in fragments]
    for w in get_all_parsed_windows():
        title = (w.get('window_title') or '').lower()
        if any(f in title for f in frags):
            return w
    return None

def _place_and_focus(w: Dict[str, Any], monitor_index: Optional[int]) -> Dict[str, Union[str, bool]]:
    wid = w.get('window_id')
    moved = True
    if wid and monitor_index is not None:
        moved = move_window_to_monitor(wid, monitor_index)
    activated = bool(wid) and activate_window(wid)
    return {"launched": False, "window_id": wid or "", "moved": moved, "activated": activated}

# ----------------- core: safe, single-instance -----------------

def get_new_window_info(
    launch_cmd: List[str],
    cwd: Optional[str],
    *,
    timeout: float = 10.0,
    poll_interval: float = 0.25,
    prefer_pid_match: bool = True,
    title_fallback: Optional[List[str]] = None,
) -> Optional[Dict[str, Any]]:
    """
    Launch once, then wait for a new window. Don't spin; obey timeout.
    Prefer matching window by spawned PID; optionally fall back to title.
    """
    before_ids = get_window_ids()
    proc = subprocess.Popen(launch_cmd, cwd=cwd)
    deadline = time.time() + timeout
    while time.time() < deadline:
        time.sleep(poll_interval)
        parsed = get_all_parsed_windows()
        pid_str = str(proc.pid)
        for w in parsed:
            wid = w.get("window_id")
            if wid not in before_ids:
                if prefer_pid_match and w.get("pid") == pid_str:
                    return w
        # Fallback direct PID
        match = next((w for w in parsed if w.get("pid") == pid_str), None)
        if match:
            return match
        if title_fallback:
            match = find_window_by_title_fragments(title_fallback)
            if match:
                return match
    return None

def ensure_single_instance_or_launch(
    *,
    path: str,
    match_strings: Optional[List[str]] = None,
    monitor_index: Union[int, str, None] = 1,
    launch_cmd: List[str],
    cwd: Optional[str] = None,
    appear_timeout_sec: float = 7.0,
    poll_interval_sec: float = 0.12,
    lock_path: Optional[str] = None,
    debounce_sec: float = 0.75,
) -> Dict[str, Union[str, bool]]:
    """
    - If window is already open: focus and raise.
    - Otherwise, launch it, wait for detection, then focus it.
    """

    # ---- lock/debounce ----
    if lock_path is None:
        lock_path = os.path.expanduser("~/.cache/abstract-ide/launch.lock")
    os.makedirs(os.path.dirname(lock_path), exist_ok=True)

    now = time.time()
    try:
        if os.path.exists(lock_path):
            with open(lock_path, "r") as f:
                data = json.load(f)
            last = float(data.get("ts", 0))
            if now - last < debounce_sec:
                existing = find_window_for_script(path)
                if existing:
                    _place_and_focus(existing, get_mon_index(monitor_index))
                    return {"launched": False, "window_id": existing.get("window_id"), "moved": True, "activated": True}
        with open(lock_path, "w") as f:
            json.dump({"ts": now}, f)
    except Exception:
        pass

    # ---- 1. Check if already running ----
    existing = find_window_for_script(path)
    if existing:
        _place_and_focus(existing, get_mon_index(monitor_index))
        focus_window(existing.get("window_title", os.path.basename(path)))
        return {"launched": False, "window_id": existing.get("window_id"), "moved": True, "activated": True}

    # ---- 2. Launch new instance ----
    try:
        proc = subprocess.Popen(launch_cmd, cwd=cwd, start_new_session=True)
        print(f"[abstract_windows] launched process PID={proc.pid}: {' '.join(launch_cmd)}")
    except Exception as e:
        print(f"[abstract_windows] launch error: {e}")
        return {"launched": False, "window_id": "", "moved": False, "activated": False}

    # ---- 3. Wait for its window to appear ----
    win = get_new_window_info(
        launch_cmd, cwd,
        timeout=appear_timeout_sec,
        poll_interval=poll_interval_sec,
        prefer_pid_match=True,
        title_fallback=(match_strings or [os.path.basename(path)]),
    )

    if not win:
        print("[abstract_windows] warning: window not detected, but process launched")
        return {"launched": True, "window_id": "", "moved": False, "activated": False}

    # ---- 4. Move + focus ----
    wid = win.get("window_id", "")
    mon_idx = get_mon_index(monitor_index)
    moved = move_window_to_monitor(wid, mon_idx) if mon_idx is not None else True
    activate_window(wid)
    focus_window(win.get("window_title", os.path.basename(path)))

    return {"launched": True, "window_id": wid, "moved": moved, "activated": True}
# ----------------- convenience: launch python with conda -----------------

def launch_python_conda_script(
    path: str,
    *,
    env_name: str = "base",
    conda_exe: str = "/home/computron/miniconda/bin/conda",
    display: str = ":0",
    monitor_index: Union[int, str, None] = 1,
) -> Dict[str, Union[str, bool]]:
    """
    Launch a python script via `conda run` (no shell sourcing), single-instance & safe.
    """
    script_abs = os.path.abspath(path)
    workdir = os.path.dirname(script_abs)

    launch_cmd = [
        conda_exe, "run", "-n", env_name, "--no-capture-output",
        "env", f"DISPLAY={display}",
        "python", script_abs,
    ]

    match_titles = [os.path.basename(script_abs)]  # light fallback
    return ensure_single_instance_or_launch(
        path=script_abs,
        match_strings=match_titles,
        monitor_index=monitor_index,   # ← integer like 1 (NOT DISPLAY)
        launch_cmd=launch_cmd,
        cwd=workdir,
        appear_timeout_sec=8.0,
        poll_interval_sec=0.15,
    )



from abstract_utilities import make_list
def get_strings_in_string(string,strings):
    for comp_string in strings:
        if comp_string.lower() in string.lower():
            return True
    return False
def find_longest_common_substring(string, comp_string):
    # Find the longest common substring
    min_len = min(len(string), len(comp_string))
    for length in range(min_len, 0, -1):
        for i in range(len(string) - length + 1):
            substring = string[i:i + length]
            if substring in comp_string:
                return substring
    return None

def get_string_confidence(string,string_remains):
    string_len = len(string)
    remains_str = ''.join(string_remains)
    remains_len = len(remains_str)
    remains_perc = 0
    confidence = 1
    if remains_len:
        remains_perc = remains_len/string_len
        confidence = 1 - remains_perc
    return confidence
def remove_substring(string, substring):
    # Remove the first occurrence of substring and return remaining parts
    idx = string.find(substring)
    if idx == -1:
        return [string]
    remains = [string[:idx], string[idx + len(substring):]]
    return [r for r in remains if r]  # Remove empty strings

def compare_strings(string, comp_string):
    results = []
    
    # Initialize with the original strings
    current_strings = [(string, comp_string)]
    
    while current_strings:
        new_strings = []
        for s, cs in current_strings:
            # Find the longest common substring
            common = find_longest_common_substring(s, cs)
            if not common:
                continue  # No common substring, move to next pair
            
            # Record the found substring and remaining parts
            string_remains = remove_substring(s, common)
            comp_string_remains = remove_substring(cs, common)
            
            # Store the result
            results.append({
                "found": common,
                "string": {"remains": string_remains},
                "comp_string": {"remains": comp_string_remains}
            })
            
            # Add remaining parts for further processing
            for sr in string_remains:
                for csr in comp_string_remains:
                    new_strings.append((sr, csr))
        
        current_strings = new_strings
    string_remains = results[-1].get('string',{}).get('remains')
    string_confidence = get_string_confidence(string,string_remains)
    comp_string_remains = results[-1].get('comp_string',{}).get('remains')
    comp_string_confidence = get_string_confidence(comp_string,comp_string_remains)
    results = {"string":string,
               "comp_string":comp_string,
               "results":results,
               "found":[result['found'] for result in results if result],
               "confidence":{
                   "string":string_confidence,
                   "comp_string":comp_string_confidence
                   }
               }
    return results

def return_best_comp_strings(string, comp_strings):
    results = []
    comp_strings = make_list(comp_strings)
    for comp_string in comp_strings:
        result = compare_strings(string, comp_string)
        results.append(result)
    
    confidence_results = None
    for i, result in enumerate(results):
        confidence = result.get('confidence', {})
        string_confidence = confidence.get('string', 0)
        comp_string_confidence = confidence.get('comp_string', 0)
        if confidence_results is None:
            #if string_confidence>.70:
                confidence_results = [string_confidence, comp_string_confidence, i]
        elif string_confidence > confidence_results[0]:
            confidence_results = [string_confidence, comp_string_confidence, i]
        elif string_confidence == confidence_results[0]:
            if comp_string_confidence > confidence_results[1]:
                confidence_results = [string_confidence, comp_string_confidence, i]
    
    return comp_strings[confidence_results[2]]




import os

def _norm(p: str) -> str:
    try:
        return os.path.abspath(os.path.realpath(p))
    except Exception:
        return os.path.abspath(p)

def find_idle_window_for_file(path: str, rows: list[dict]) -> dict | None:
    """Find an open IDLE editor window that has file_path opened."""
    target = _norm(path)
    base   = os.path.basename(target)

    for w in rows:
        sig = w.get("program_signature") or {}
        args = sig.get("args") or []
        mod  = (sig.get("module") or "").lower()

        # Best: IDLE launched via -m idlelib.idle with the target path in args
        if mod == "idlelib.idle" and any(_norm(a) == target for a in args if a.endswith(".py")):
            return w

        # Fallback: window title check (IDLE usually shows "name.py - /full/path (...)" )
        title = (w.get("window_title") or "").lower()
        if base.lower() in title and os.path.dirname(target).lower() in title:
            return w
    return None
